/* Copyright (C) 2006-2017 Michel Gutierrez <mig@downloadhelper.net>
 * All Rights Reserved
 * 
 * If you are interested in methods used in this project, follow the
 * open source project at https://github.com/mi-g/fang
 */

const $chrome = require("chrome");
const Cc = $chrome.Cc;
const Ci = $chrome.Ci;
const Cu = $chrome.Cu;
const CC = $chrome.CC;

const merge = require('sdk/util/object').merge;
const timers = require("sdk/timers");
const simplePrefs = require("sdk/simple-prefs");
const _ = require("sdk/l10n").get;
/* @+chrome
const Request = require("sdk/request").Request;
   @- */

const hits = require("./hits");
const utils = require("./utils");
const log = require("./log");
const m3u8 = require("./m3u8");
const hls = require("./hls");
const dash = require("./dash");
const f4f = require("./f4f");
const amf = require("./amf");

const JSON_MASTER_PATTERN = new RegExp("^https?://.*/master\\.json");
const DASH_XML_CT_PATTERN = new RegExp("dash.*mpd");
const M3U8_PATTERN = new RegExp("^https?://.*\\.m3u8(?:\\?|$)");
const JSON_HLS_PS_PATTERN = new RegExp("^https?://api\\.periscope\\.tv/api/v2/getAccessPublic");
const EMPTY_LINE_PATTERN = new RegExp("^\\s*$");

const F4F_PATTERN = new RegExp("^https?://.*\\.f4m(?:\\?|$)");
const F4F_PATTERN_FRAG = new RegExp("^(https?://.*)Seg1\\-Frag([0-9]+)(\\?.*)?$");

/* @+firefox */
const BinaryInputStream = CC('@mozilla.org/binaryinputstream;1', 'nsIBinaryInputStream', 'setInputStream');
const BinaryOutputStream = CC('@mozilla.org/binaryoutputstream;1', 'nsIBinaryOutputStream', 'setOutputStream');
const StorageStream = CC('@mozilla.org/storagestream;1', 'nsIStorageStream', 'init');
/* @- */

exports.networkHook = function(channel,meta) {
	if(!simplePrefs.prefs["chunks.enabled"])
		return null;
	var probe = null;
	if(simplePrefs.prefs["dash.enabled"]) {
		if(JSON_MASTER_PATTERN.test(channel.name))
			probe = new Dash(channel.name,"json",meta);
		else if(meta.contentType && DASH_XML_CT_PATTERN.test(meta.contentType.toLowerCase()))
			probe = new Dash(channel.name,"xml",meta);
	}
	if(simplePrefs.prefs["hls.enabled"]) {
		if(M3U8_PATTERN.test(channel.name))
			probe = new Hls(channel.name);
		else if(JSON_HLS_PS_PATTERN.test(channel.name))
			probe = new Hls(channel.name,"json");		
		else if(meta.contentType && meta.contentType.toLowerCase().indexOf("mpegurl")>=0)
			probe = new Hls(channel.name);
	}
	if(!probe && simplePrefs.prefs["f4f.enabled"]) {
		probe = F4f.getProbe(channel);
	}
	if(probe) {
		if(!probe.skipManifest) {
/* @+firefox */
			var tracChannel = channel.QueryInterface(Ci.nsITraceableChannel);
			probe.originalListener = tracChannel.setNewListener(probe);
/* @- */
/* @+chrome
			Request({
				url: channel.name,
				onComplete: function(response) {
					probe.handleManifest(response.text);
					probe.checkReady();
				},
			}).get();

   @- */
		}
		return probe;
	}
	return null;
}

function Probe() {
}

Probe.prototype = {
	init: function(type) {
		this.type = type;
		this.receivedChunks = [];
	},
	handleHit: function(hitData) {
		this.hitData = hitData;
		this.checkReady();
	},
/* @+firefox */
	onDataAvailable: function(aRequest, aContext, aInputStream, aOffset, aCount) {
		var iStream = new BinaryInputStream(aInputStream);
		var sStream = new StorageStream(8192, aCount, null);
		var oStream = new BinaryOutputStream(sStream.getOutputStream(0));
		var data = iStream.readBytes(aCount);
		this.receivedChunks.push(data);
		oStream.writeBytes(data, aCount);
		this.originalListener.onDataAvailable(aRequest, aContext, sStream.newInputStream(0), aOffset, aCount);
	},
	onStartRequest: function(aRequest, aContext) {
		this.originalListener.onStartRequest(aRequest, aContext);
	},
	onStopRequest: function(aRequest, aContext, aStatusCode) {
		var body = this.receivedChunks.join("");
		delete this.receivedChunks;
		try {
			this.handleManifest(body);
			this.checkReady();
		} catch(e) {
			console.error("chunks manifest error",e.message);
		}
		this.originalListener.onStopRequest(aRequest, aContext, aStatusCode);

	},
/* @- */
	checkReady: function() {},
	handleManifest: function() {},
	handle: function() {},
};

function Dash(url,format,meta) {
	this.format = format;
	this.manifestUrl = url;
	this.meta = meta;
	this.init("dash");
}

Dash.prototype = new Probe();

Dash.prototype.handleManifest = function(body) {
	try {
		if(this.format=="json") {
			var mpd = JSON.parse(body);
			if(mpd && Array.isArray(mpd.video) && mpd.video.length>0 && Array.isArray(mpd.video[0].segments) && mpd.video[0].segments.length>0)
				this.mpd = mpd;
		} else if(this.format=="xml") {
/* @+firefox */
			this.handler = require("./tbvws").dashManifest(this.manifestUrl,body,this.meta);
/* @- */
		}
	} catch(e) { 
		console.error("Error parsing DASH manifest",e.message || e); 
	}
}
Dash.prototype.checkReady = function() {
	if(this.hitData && (this.mpd || this.handler))
		this.handle();
}

Dash.prototype.pickAudioMpd = function() {
	var prefs = [{
		field: "codecs",
		pref: "mp4a.40.2"
	},{
		field: "format",
		pref: "mp42",
	},{
		field: "mime_type",
		pref: "audio/mp4"
	},{
		field: "channels",
		pref: "@max"
	},{
		field: "bitrate",
		pref: "@max"
	},{
		field: "sample_rate",
		pref: "@max"
	}];
	var audios = [].concat(this.mpd.audio);
	audios.sort(function(a,b) {
		for(var i=0;i<prefs.length;i++) {
			var crit = prefs[i];
			if(a[crit.field]!=b[crit.field]) {
				if(crit.pref=="@max")
					return b[crit.field] - a[crit.field];
				if(a[crit.field]==crit.pref)
					return -1;
				if(b[crit.field]==crit.pref)
					return 1;
			}
		}
		return 0;
	});
	return audios[0];
}

Dash.prototype.handle = function() {
	var hash = utils.md5(this.hitData.url);
	var self = this;
	if(this.handler)
		this.handler(this.hitData);
	else if(this.mpd) {
/* @+firefox */
		Cu.import("resource://gre/modules/NetUtil.jsm");
/* @- */
		var hasAudio = self.mpd.audio && self.mpd.audio.length>0 && 
			Array.isArray(self.mpd.audio[0].segments) && self.mpd.audio[0].segments.length>0;
		var hasVideo = self.mpd.video && self.mpd.video.length>0 && 
			Array.isArray(self.mpd.video[0].segments) && self.mpd.video[0].segments.length>0;
		var streamLoop = this.mpd.video; 
		var extraStream = this.mpd.audio;
		if(!hasVideo || (hasAudio && simplePrefs.prefs["dash.on-adp"]=="audio")) {
			streamLoop = this.mpd.audio;
			extraStream = null;
		} else if(!hasAudio || (hasVideo && simplePrefs.prefs["dash.on-adp"]=="video"))
			extraStream = null;
		streamLoop.forEach(function(mpd,index) {
			var mpdExtra = {};
/* @+firefox */
			if(extraStream) {
				mpdExtra.chunked = "dash-adp",
				mpdExtra.audioMpd = self.pickAudioMpd();
				mpdExtra.videoMpd = mpd;
				mpdExtra.audioUrl = NetUtil.newURI(self.hitData.url).resolve("dash-audio.mp4");
				mpdExtra.videoUrl = NetUtil.newURI(self.hitData.url).resolve("dash-video.mp4");
				mpdExtra.url = undefined;
			} else
/* @- */
				mpdExtra._mpd = mpd;
			var hitData = merge({},self.hitData,{
				id: "dash:"+hash+"-"+index,
				"extension": "mp4",
				length: null,
				chunked: "dash",
				descrPrefix: _("dash-streaming"),
				group: "grp-"+hash,
			},mpdExtra);

			hitData._mpdCommonBaseUrl = self.mpd.base_url;
			if(mpd.width && mpd.height)
				hitData.size = mpd.width + "x" + mpd.height;
			if(mpd.duration)
				hitData.duration = Math.round(mpd.duration);
			hits.newData(hitData);
		});
	}
}

function Hls(url,masterFormat) {
	this.init("hls");
	this.masterFormat = masterFormat || "m3u8"; 
	this.mediaUrl = url;
}

Hls.prototype = new Probe();

Hls.prototype.handleHit = function(hitData) {
/* @+chrome
	delete hitData.length;
   @- */
	hitData.group = "grp-" + utils.md5(hitData.url);
	if(simplePrefs.prefs["m3u8.details"])
		hitData.masterManifest = hitData.url;
	Probe.prototype.handleHit.call(this,hitData);
}

Hls.prototype.handleManifest = function(body) {
	var manifest = null;
	if(this.masterFormat=="m3u8")
		manifest = m3u8.get(body,this.mediaUrl);
	else if(this.masterFormat=="json")
		manifest = m3u8.getPsJson(body,this.mediaUrl);
	if(manifest) {
		if(manifest.isMaster())
			this.master = manifest;
		else if(manifest.isMedia())
			this.media = manifest;
	}
}
Hls.prototype.checkReady = function() {
	if(this.hitData && (this.master || this.media))
		this.handle();
}
Hls.prototype.handle = function() {
	if(this.master)
		hls.handleMaster(this.master,this.hitData);
	else if(this.media)
		hls.handleMedia(this.media,this.hitData,this.mediaUrl);
}

function F4f(url) {
	this.init("f4f");
	this.f4fUrl = url;
/* @+firefox */
	Cu.import("resource://gre/modules/NetUtil.jsm");
/* @- */
	if(simplePrefs.prefs['f4f.frag-index']) {
/* @+firefox */
		var url0 = NetUtil.newURI(url);
		var rootUrl = /^([^\?]*?)([^\/]*\?.*|[^\/]*)$/.exec(url0.prePath + url0.path)[1];
/* @- */
/* @+chrome
		var url0 = new URL(url);
		var rootUrl = /^([^\?]*?)([^\/]*\?.*|[^\/]*)$/.exec(url0.protocol+"//"+url0.host+url0.pathname)[1];
   @- */
		F4f.waitingForFrag[rootUrl] = this;
	}
}

F4f.getProbe = function(channel) {
	if(F4F_PATTERN.test(channel.name)) {
		return new F4f(channel.name);
	} else if(simplePrefs.prefs['f4f.frag-index']) {
		var m = F4F_PATTERN_FRAG.exec(channel.name);
		if(m) {
/* @+firefox */
			Cu.import("resource://gre/modules/NetUtil.jsm");
/* @- */
			var rootUrl = channel.name;
			while(true) {
/* @+firefox */
				var url0 = NetUtil.newURI(rootUrl);
				rootUrl = /^(https?:\/\/[^\/]+\/.*?)(\/|[^\/]+)$/.exec(url0.prePath + url0.path);
/* @- */
/* @+chrome
				var url0 = new URL(rootUrl);
				var rootUrl = /^(https?:\/\/[^\/]+\/.*?)(\/|[^\/]+)$/.exec(url0.protocol+"//"+url0.host+url0.pathname);
   @- */
				if(!rootUrl) {
					// manifest was missed, construct the probe from the data fragment
					var probe = new F4f(channel.name);
					//probe.startFrag = parseInt(m[2]);
					probe.startFrag = 1;
					probe.postFrag = m[3];
					probe.skipManifest = true;
					probe.rootUrl = m[1];
					probe.medias = {};
					probe.medias[m[1]] = {
						urlHint: m[1],
					}
					return probe;
				}
				rootUrl = rootUrl[1];
				var probe1 = F4f.waitingForFrag[rootUrl];
				if(probe1)
					delete F4f.waitingForFrag[rootUrl];
				else
					for(var pid in F4f.waitingForFrag) {
						var probe0 = F4f.waitingForFrag[pid];
						for(var mid in probe0.medias) {
							var media = probe0.medias[mid];
							if(media.urlHint==rootUrl) {
								delete F4f.waitingForFrag[pid];
								probe1 = probe0;
								break;
							}
						}
						if(probe1)
							break;
					}
				if(probe1) {
					var hit = hits.getF4fHitDataByUrlHint(m[1]);
					if(hit)
						return null;
					if(!probe1.startFrag)
						probe1.startFrag = parseInt(m[2]);
					if(!probe1.postFrag)
						probe1.postFrag = m[3];
					probe1.checkReady();
					return null;
				}
			}
		}
		return null;
	}	
}

F4f.waitingForFrag = {};

F4f.prototype = new Probe();

F4f.prototype.checkReady = function() {
	if(this.hitData && this.medias && (this.startFrag || !simplePrefs.prefs['f4f.frag-index']))
		this.handle();
}

F4f.prototype.handleManifest = function(body) {
/* @+firefox */
	var parser = Cc["@mozilla.org/xmlextras/domparser;1"].createInstance(Ci.nsIDOMParser);
	var doc = parser.parseFromString(body, "application/xml");
	Cu.import("resource://gre/modules/NetUtil.jsm");
/* @- */
/* @+chrome
	var doc = new DOMParser().parseFromString(body, "application/xml");
   @- */
	var durationElement = doc.querySelector("duration");
	if(durationElement)
		this.duration = parseInt(durationElement.firstChild.nodeValue);
	var bis = doc.querySelectorAll("bootstrapInfo");
	this.medias = {};
	for(var i=0;i<bis.length;i++) {
		var bi = bis.item(i);
		var id = bi.getAttribute("id");
		var data = (bi.firstChild && utils.toByteArray(bi.firstChild.nodeValue)) || null;
		this.medias[id] = {
			bootstrap: data,
		}
	}
	var mes = doc.querySelectorAll("media");
	for(var i=0;i<mes.length;i++) {
		var m = mes.item(i);
		var mid = m.getAttribute("bootstrapInfoId");
		var media = this.medias[mid];
		if(media) {
			media.bitrate = parseInt(m.getAttribute("bitrate"))*1000;
/* @+firefox */
			media.urlHint = NetUtil.newURI(this.f4fUrl).resolve(m.getAttribute("url"));
/* @- */
/* @+chrome
			media.urlHint = new URL(m.getAttribute("url"),this.f4fUrl).href;
   @- */
			var metaElement = m.querySelector("metadata");
			if(metaElement) {
				var metaBytes = (metaElement.firstChild && utils.toByteArray(metaElement.firstChild.nodeValue)) || null;
				if(metaBytes) {
					var metaVars = amf.decode(metaBytes);
					if(metaVars && metaVars.length>=2 && metaVars[0]=="onMetaData")
						media.meta = metaVars[1];
				}
			}
		}
	}
}

F4f.prototype.handle = function() {
	var group = "grp-" + utils.md5(this.rootUrl || this.hitData.url);
	for(var id in this.medias) {
		var media = this.medias[id];
		if(this.skipManifest) {
			var oldHitData = hits.getF4fHitDataByUrlHint(media.urlHint);
			if(oldHitData)
				continue;
		}
		this.hitData.url = this.rootUrl || this.hitData.url;
		var hash = utils.md5(this.hitData.url+id);
		var hitData = merge({},this.hitData,{
			id: "f4f:"+hash,
            "extension": "flv",
			bitrate: media.bitrate,
			_media: media,
			length: null,
			chunked: "f4f",
			descrPrefix: _("f4f-streaming"),
			startFrag: this.startFrag,
			postFrag: this.postFrag || "",
			group: group,
		});
		if(this.duration)
			hitData.duration = this.duration;
		if(media.meta) {
			var meta = media.meta;
			if(meta.duration)
				hitData.duration = Math.round(meta.duration);
			if(meta.width && meta.height)
				hitData.size = meta.width + "x" + meta.height;
			if(meta.filesize)
				hitData.length = meta.filesize;
			if(meta.framerate)
				hitData.fps = meta.framerate;
		}
		hits.newData(hitData);
	}
}

exports.download = function(action,specs,successFn,errorFn,progressFn) {
	specs.ignoreSpecs = true;
	var chunkSet = null;
	switch(action.hit.data.chunked) {
	case "hls":
		chunkSet = hls.getChunkSet(action.hit.data);
		break;
	case "dash":
		chunkSet = new dash.DashChunkset();
		chunkSet.init(action.hit.data);
		break;
/* @+firefox */
	case "dash-adp":
		DownloadAdp(action,specs,successFn,errorFn,progressFn);
		return;
/* @- */
	case "f4f":
		chunkSet = new f4f.F4fChunkset();
		chunkSet.init(action.hit.data);
		break;
	}
	if(!chunkSet) {
		log.error("Requested download of chunked stream, but no chunkset found");
		action.cleanup();
		action.hit.setCurrentAction(null);
		action.notifyRunning(false);
		return;
	}
	chunkSet.download(action,specs,successFn,errorFn,progressFn);
}

/* @+firefox */
function DownloadAdp(action,specs,successFn,errorFn,progressFn) {
	Cu.import("resource://gre/modules/NetUtil.jsm");
	var hitData = action.hit.data;
	var percents = {
		audio: 0,
		video: 0,
	};
	var successes = {
		audio: false,
		video: false,
	}
	var errorReported = false;
	function ErrorFn(err) {
		if(errorReported)
			return;
		errorReported = true;
		log.error(err.message);
		action.cleanup();
		action.hit.setCurrentAction(null);
		action.notifyRunning(false);
		errorFn(err);
	}
	var chunkSets = {};
	var aborted = false;
	["audio","video"].forEach(function(which) {
		function Success() {
			successes[which] = true;
			if(successes.audio && successes.video)
				successFn();
		}
		function Progress(percent) {
			percents[which] = percent;
			if(!aborted)
				progressFn(Math.round((percents.audio+percents.video)/2));
		}
		try {
			var chunkSet = new dash.DashChunkset();
			var mpd = hitData[which+"Mpd"];

			function StartDownload(initSegmentData) {
				chunkSets[which] = {
					chunkset: chunkSet,
					target: hitData["_"+which+"Target"]
				}
				chunkSet.init({});

				var baseUrl = NetUtil.newURI(hitData.url || hitData[which+"Url"]);
				if(hitData._mpdCommonBaseUrl)
					baseUrl = NetUtil.newURI(baseUrl.resolve(hitData._mpdCommonBaseUrl));
				if(mpd.base_url)
					baseUrl = NetUtil.newURI(baseUrl.resolve(mpd.base_url));

				chunkSet.downloadFile(hitData["_"+which+"Target"],{
					init_segment: initSegmentData,
					segments: mpd.segments,
				},baseUrl.spec,Success,ErrorFn,Progress);
			}

			if(typeof mpd.init_segment == "string")
				StartDownload(utils.toByteArray(mpd.init_segment))
			else if(typeof mpd.init_segment == "object")
				StartDownload(mpd.init_segment)
			else {
				var initSegmentUrl = NetUtil.newURI(hitData[which+"Url"]).resolve(mpd.init_segment_url);
				utils.DownloadToByteArray(initSegmentUrl,hitData.headers,hitData.isPrivate,false,
					function(err,data) {
						if(err)
							return Error(err.message);
						StartDownload(data);
					});
			}
		} catch(e) {
			ErrorFn(new Error("Dash ADP: "+e.message));
		}
	});
	action.hit.abortFn = function() {
		aborted = true;
		for(var which in chunkSets)
			chunkSets[which].chunkset.actionAbortFn(chunkSets[which].target+".part");
		action.cleanup();
		action.hit.setCurrentAction(null);
		action.notifyRunning(false);
		errorFn(_("download-canceled"));
	}
}
/* @- */
